# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import sys

try:
    import cPickle as pickle
except ImportError:
    import pickle

# shamelessly ripped from https://github.com/kennethreitz/requests/blob/master/requests/compat.py
# Syntax sugar.
_ver = sys.version_info
is_py2 = _ver[0] == 2
is_py3 = _ver[0] == 3

if is_py2:
    basestring = basestring
    unicode = unicode
    range = xrange # range is now always an iterator
    input = raw_input
    cmp = cmp

    import Queue as queue

    # using _* makes __all__ not export this automatically, so you have to
    # explicitly import this (from datatypes.compat import _thread), the reason
    # is because the thread module is lowkey deprecated in favor of the
    # threading module
    #
    # https://stackoverflow.com/questions/1141047/why-was-the-thread-module-renamed-to-thread-in-python-3-x
    #
    # I've chosen to keep the _thread syntax to prod me into finding a better
    # way to do something, but I still have access to it explicitely if I need it
    import thread as _thread

    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO

    from SimpleHTTPServer import SimpleHTTPRequestHandler
    from BaseHTTPServer import (
        BaseHTTPRequestHandler,
        HTTPServer,
    )
    import SocketServer as socketserver
    import Cookie as cookies
    import __builtin__ as builtins

    from urllib import urlencode
    from urllib2 import (
        Request,
        urlopen,
        URLError,
        HTTPError,
    )
    import urlparse as parse

    import itertools
    zip = itertools.izip
    zip_longest = itertools.izip_longest

    from collections import (
        Iterable,
        Mapping,
        Callable,
        MutableSequence,
        Sequence,
    )


    from HTMLParser import HTMLParser
    #unescape = HTMLParser().unescape

    import cgi
    class html(object):
        @classmethod
        def escape(cls, *args, **kwargs):
            return cgi.escape(*args, **kwargs)

        @classmethod
        def unescape(cls, *args, **kwargs):
            return HTMLParser().unescape(*args, **kwargs)

    # NOTE -- getargspace isn't a full mapping of getfullargspec
    from inspect import getargspec as getfullargspec


    exec("""def reraise(exception_class, e, traceback=None):
        try:
            raise exception_class, e, traceback
        finally:
            traceback = None
    """)


else:
    basestring = (str, bytes)
    unicode = str
    long = int
    input = input

    import queue
    import _thread
    from io import StringIO
    from http.server import (
        HTTPServer,
        SimpleHTTPRequestHandler,
        BaseHTTPRequestHandler,
    )
    import socketserver
    from http import cookies
    import builtins

    from html.parser import HTMLParser
    from urllib import parse
    from urllib.parse import urlencode
    from urllib.request import Request, urlopen
    from urllib.error import URLError, HTTPError
    try:
        import html
        #from html import unescape
    except ImportError:
        import cgi
        class html(object):
            @classmethod
            def escape(cls, *args, **kwargs):
                return cgi.escape(*args, **kwargs)

            @classmethod
            def unescape(cls, *args, **kwargs):
                return HTMLParser().unescape(*args, **kwargs)
        #unescape = HTMLParser.unescape

    from inspect import getfullargspec
    from itertools import zip_longest

    from collections.abc import (
        Iterable,
        Mapping,
        Callable,
        MutableSequence,
        Sequence,
    )


    # ripped from six https://github.com/benjaminp/six
    def reraise(exception_class, e, traceback=None):
        """the 3 params correspond to the return value of sys.exc_info()

        https://docs.python.org/3/library/sys.html#sys.exc_info

        :param exception_class: BaseException, the class of the exception to reraise
        :param e: BaseException instance, the actual exception instance
        :param traceback: traceback, the stack trace
        """
        try:
            e = exception_class("" if e is None else e)
            if e.__traceback__ is not traceback:
                raise e.with_traceback(traceback)
            raise e
        finally:
            e = None
            traceback = None


    # py3 has no cmp function for some strange reason
    # https://codegolf.stackexchange.com/a/49779
    def cmp(a, b):
        return (a > b) - (a < b)


Str = unicode if is_py2 else str
Bytes = str if is_py2 else bytes


# TODO using reraise

#             if py_2:
#                 #raise error_info[0].__class__, error_info[0], error_info[1][2]
#                 reraise(*error_info)
#                 #raise error_info[0].__class__, error_info[1], error_info[2]
# 
#             elif py_3:
#                 #e, exc_info = error_info
#                 #et, ei, tb = exc_info
# 
#                 reraise(*error_info)
#                 #et, ei, tb = error_info
#                 #raise ei.with_traceback(tb)


# if not error_info:
#                     exc_info = sys.exc_info()
#                     #raise e.__class__, e, exc_info[2]
#                     #self.error_info = (e, exc_info)
#                     self.error_info = exc_info
# 
# if error_info:
# 
#             reraise(*error_info)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import re

from .compat import *
from .string import String


class Shorten(object):
    """Converts an integer to a base58-like string and vice versa

    This code was used in Undrip for its link shortener and was originally written
    by Cahlan Sharp and Ryan Johnson (I believe) and based off of this:

        https://stackoverflow.com/a/2549514/5006

    any ambiguous characters were removed from BASE_LIST (notice no I, 1, or l) to
    make it easier for humans to copy the url by hand

    this is a version of base58 that takes strings:
        https://github.com/keis/base58
    """
    BASE_LIST = '23456789' + 'abcdefghijkmnpqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ' + '_-'
    BASE_DICT = dict((c, i) for i, c in enumerate(BASE_LIST))

    def __new__(cls, val):
        if isinstance(val, basestring):
            ret = cls.decode(val)

        elif isinstance(val, int):
            ret = cls.encode(val)

        else:
            raise TypeError("Unknown type {}".format(type(val)))

        return ret

    @classmethod
    def encode(cls, integer, encoding=""):
        base = cls.BASE_LIST
        length = len(base)
        ret = ''
        while integer != 0:
            ret = base[integer % length] + ret
            integer //= length
        return String(ret)

    @classmethod
    def decode(cls, s, encoding=""):
        reverse_base = cls.BASE_DICT
        length = len(reverse_base)
        ret = 0
        for i, c in enumerate(s[::-1]):
            ret += (length ** i) * reverse_base[c]
        return ret


class Integer(int):
    def hex(self, length=0, prefix=""):
        format_str = "{}{{:0>{}X}}".format(prefix, length) if length else "{}{{:X}}".format(prefix)
        return format_str.format(self)

    def binary(self, length=0, prefix=""):
        format_str = "{}{{:0>{}b}}".format(prefix, length) if length else "{}{{:b}}".format(prefix)
        return format_str.format(self)

    def range(self, start=0):
        """iterate from 0 to self, if self is negative it goes from self to 0

        I honestly have no idea why this is useful but it was in my notes for this
        module, so here it is
        """
        if self > start:
            r = range(0, self)
        else:
            r = range(self, start)

        for i in r:
            yield i

    def __new__(cls, v, base=None):
        if not isinstance(v, int):
            if base is None and isinstance(v, basestring):
                # NOTE: if you have an ambiguous value like "0001" or "0010" this
                # will assume it is binary
                if re.match(r"^[-+]?0b", v) or re.search(r"[01]+$", v):
                    # https://stackoverflow.com/questions/8928240/convert-base-2-binary-number-string-to-int
                    base = 2

                elif re.match(r"^[-+]?0x", v) or re.search(r"[a-fA-F0-9]+$", v):
                    base = 16

            v = int(v, base)

        return super(Integer, cls).__new__(cls, v)


class Hex(Integer):
    def __new__(cls, v):
        return super(Hex, cls).__new__(cls, v, 16)


class Binary(Integer):
    def __new__(cls, v):
        return super(Binary, cls).__new__(cls, v, 2)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import os
import codecs
import csv
import logging

from .compat import *
from . import environ
from .string import String, ByteString
from .utils import cbany


logger = logging.getLogger(__name__)


class CSV(object):
    """Easily read/write the rows of a csv file
    :Example:
        # read a csv file
        c = CSV("/path/to/read/file.csv")
        for row in c:
            print(row)
        # write a csv file
        c = CSV("/path/to/write/file.csv")
        with c:
            for row in rows:
                c.add(row)
    """
    reader_class = csv.DictReader
    """the class that will be used by .create_reader()"""

    writer_class = csv.DictWriter
    """the class that will be used by .create_writer()"""

    rows_class = list
    """the class used in .rows()"""

    reader_row_class = None
    """You can set this to a class and rows returned from the default .normalize_reader_row()
    will be this type"""

    writer_row_class = None
    """You can set this to a class and rows returned from the default .normalize_writer_row()
    will be this type, this class should act like a dict unless you also change .writer_class"""

    class ContinueError(Exception):
        """Can be thrown to have CSV skip the current row"""
        pass

    def __init__(self, path, fieldnames=None, encoding="", **kwargs):
        """Create the csv instance
        :param path: string, the path to the csv file that will be read/written
        :param fieldnames: list, the fieldnames, when writing, if this is omitted
            then the keys of the first row dictionary passed to .add() will be used
            for the fieldnames. If omitted when reading then the first line of the
            csv file will be used for the fieldnames
        :param encoding: string, what character encoing to use
        """
        self.path = path
        self.fieldnames = self.normalize_fieldnames(fieldnames or [])
        self.writer = None
        self.reader = None
        self.context_depth = 0

        if not encoding:
            encoding = environ.ENCODING
        self.encoding = encoding

        cls = type(self)
        for k, v in kwargs.items():
            if hasattr(cls, k):
                setattr(self, k, v)

    def __iter__(self):
        logger.debug("Reading csv file: {}".format(self.path))

        with self.open() as f:
            first_row = True
            self.reader = self.create_reader(f)
            for row in self.reader:
                try:
                    row = self.normalize_reader_row(row)
                    if row:
                        if self.fieldnames and first_row:
                            # if you've passed in fieldnames then it will return
                            # fieldnames mapped to fieldnames as the first row,
                            # this will catch that and ignore it
                            #
                            # if we pass in fieldnames then DictReader won't use the first
                            # row as fieldnames, so we need to check to make sure the first
                            # row isn't a field_name: field_name mapping
                            if cbany(lambda r: r[0] != r[1], row.items()):
                                yield row

                        else:
                            yield row

                except self.ContinueError:
                    pass

                finally:
                    first_row = False

    def open(self, mode=""):
        """Mainly an internal method used for opening the file pointers needed for
        reading and writing
        :param mode: string, the open mode
        :returns: file pointer
        """
        if is_py2:
            if not mode:
                mode = "rb"
            return open(self.path, mode=mode)

        else:
            if not mode:
                mode = "r"
            return codecs.open(self.path, encoding=self.encoding, mode=mode)

    def __enter__(self):
        """Enables with context manager for writing"""
        self.context_depth += 1
        logger.debug("Writing ({}) csv file: {}".format(self.context_depth, self.path))
        if not self.writer:
            f = self.open("ab+")
            self.writer = self.create_writer(f)
        return self

    def __exit__(self, exception_type, exception_val, trace):
        self.context_depth -= 1
        if self.context_depth <= 0:
            self.writer.f.close()
            self.writer = None
            self.context_depth = 0

    def create_writer(self, f, **kwargs):
        kwargs.setdefault("dialect", csv.excel)
        kwargs.setdefault("restval", "")
        kwargs.setdefault("extrasaction", "ignore")
        kwargs.setdefault("quoting", csv.QUOTE_MINIMAL)
        kwargs.setdefault("fieldnames", self.fieldnames)
        #kwargs.setdefault("fieldnames", ["foo", "bar"])

        # from testdata CSVpath code:
        # in order to make unicode csvs work we are going to do a round about
        # thing where we write to a string buffer and then pull that out and write
        # it to the file, this is the only way I can make utf-8 work
        queue = self.normalize_writer_file(f)
        writer = self.writer_class(queue, **kwargs)
        writer.f = f
        writer.queue = queue
        writer.has_header = True if os.path.getsize(self.path) > 0 else False
        return writer

    def create_reader(self, f, **kwargs):
        """create a csv reader, this exists to make it easy to customize functionality,
        for example, you might have a csv file that doesn't have column headers, so you
        can override this method to pass in the column names, etc.
        :param f: io object, usually a file path opened with open()
        :returns: csv.Reader instance or something that acts like a built-in csv.Reader
            instance
        """
        kwargs.setdefault("fieldnames", self.fieldnames or None)
        kwargs.setdefault("dialect", csv.excel)

        f = self.normalize_reader_file(f)
        reader = self.reader_class(f, **kwargs)
        reader.f = f
        return reader

    def normalize_writer_file(self, f):
        queue = StringIO()
        return queue

    def normalize_reader_file(self, f):
        # https://stackoverflow.com/a/30031962/5006
        class FileWrapper(object):
            def __init__(self, f):
                self.f = f
                self.last_line = "" # will contain raw CSV row

            def __iter__(self):
                return self

            def __next__(self):
                self.last_line = next(self.f)
                return self.last_line
            next = __next__ # needed for py2? Python didn't consider it an iterator without next()

        return FileWrapper(f)

    def normalize_reader_row(self, row):
        """prepare row for reading, meant to be overridden in child classes if needed"""
        if is_py2:
            row = {String(k): String(v) for k, v in row.items()}

        if self.reader_row_class:
            row = self.reader_row_class(row)

        return row

    def normalize_writer_row(self, row):
        """prepare row for writing, meant to be overridden in child classes if needed"""

        row = {String(k): ByteString(v) for k, v in row.items()}

        if self.writer_row_class:
            row = self.writer_row_class(row)

        return row

    def add(self, row):
        with self:
            writer = self.writer

            if not writer.has_header:
                if not self.fieldnames:
                    self.fieldnames = self.normalize_fieldnames(row.keys())
                writer.fieldnames = self.fieldnames
                logger.debug("Writing fieldnames: {}".format(", ".join(self.fieldnames)))
                writer.writeheader()
                writer.has_header = True

            try:
                row = self.normalize_writer_row(row)
                if row:
                    writer.writerow(row)
                    data = writer.queue.getvalue()
                    writer.f.write(data)
                    writer.queue.truncate(0)
                    writer.queue.seek(0)

            except self.ContinueError:
                pass

    def append(self, rows):
        """Add all the rows to the end of the csv file

        :param rows: list, all the rows
        """
        with self:
            for row in rows:
                self.add(row)

    def normalize_fieldnames(self, fieldnames):
        return list(map(String, fieldnames))

    def find_fieldnames(self):
        """attempt to get the field names from the first line in the csv file"""
        with self.open() as f:
            reader = self.create_reader(f)
            return list(map(String, reader.fieldnames))

    def rows(self):
        """Return all the rows as a list"""
        return self.rows_class(self.__iter__())

    def clear(self):
        """clear the csv file"""
        with self.open("wb") as f:
            f.truncate(0)

    def __len__(self):
        """Returns how many rows are in this csv, this actually goes through all the
        rows, so this is not a resource light method"""
        count = 0
        for r in self:
            count += 1
        return count




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import copy
from io import IOBase

from .compat import *


class Deepcopy(object):
    """deep copy an object trying to anticipate and handle any errors

    The original version of this class comes from endpoints.utils and is used in
    the http.Http.__deepcopy__() method.

    You would traditionally use this in a class by adding these two methods:

        def copy(self):
            # handy wrapper around the deepcopy stuff
            return copy.deepcopy(self)

        def __deepcopy__(self, memodict=None):
            dc = Deepcopy()
            return dc.copy(self, memodict)

    This follows the rules of pickle for creating a new instance, so you can use
    methods like __getnewargs_ex__ to get the arguments for creating a new instance
    and __getstate__() and __setstate__() to handle migrating data from the old instance
    to the new instance

    https://docs.python.org/3/library/copy.html
    """
    def __init__(self, **kwargs):
        """
        :param **kwargs:
            ignore_keys: set, if you don't want to include certain keys in the copy
                then you can pass them in this named argument
            ignore_private: boolean, if you don't want to copy any properties that
                begin with an underscore (eg, _foo) then set this to True
            memodict: the memodict used in deepcopy
        """
        self.ignore_keys = set(kwargs.pop("ignore_keys", []))
        self.memodict = kwargs.pop("memodict", kwargs.pop("memo", {}))
        self.ignore_private = kwargs.pop("ignore_private", False)

    def copy(self, val, memodict=None, **kwargs):
        if memodict:
            memo = dict(self.memodict)
            memo.update(memodict or {})
        else:
            memo = self.memodict

        ignore_keys = kwargs.get("ignore_keys", self.ignore_keys)
        ignore_private = kwargs.get("ignore_keys", self.ignore_private)

        if hasattr(val, "__dict__"):
            instance = self.create_instance(val)
            d = self.get_state(val)

            instance_d = {}
            for k in d.keys():
                if k in ignore_keys:
                    continue

                if ignore_private and k.startswith("_"):
                    continue

                if k in memo:
                    instance_d[k] = memo[k]

                else:
                    instance_d[k] = self._copy(d[k], memo)

            self.set_state(instance, instance_d)

        else:
            instance = self._copy(val, memo)

        return instance

    def _copy(cls, val, memodict):
        ret = val
        # try a deepcopy
        try:
            ret = copy.deepcopy(val, memodict)

        except (AttributeError, TypeError) as e:
            # if the deepcopy failed a shallow copy is better than nothing
            try:
                ret = copy.copy(val)

            except (TypeError, RuntimeError):
                # we can do nothing so just pass it straight through
                # RuntimeError is because py3 can throw a RecursionError, py2
                # will throw a TypeError
                pass

        return ret

    def create_instance(self, val):
        """If the passed in .copy val is a class, this will create a new instance
        of val using the pickle rules:

            py3: https://docs.python.org/3/library/pickle.html#module-pickle
            py2: https://docs.python.org/2/library/pickle.html#module-pickle

        This will try for py3 methods first and then fall back to py2 rules
        """
        instance_class = self.get_instance_type(val)
        args, kwargs = self.get_instance_args(val)
        return instance_class(*args, **kwargs)

    def get_instance_type(self, val):
        """get the class/type of val"""
        return type(val)

    def get_instance_args(self, val):
        """get the arguments that val would need to be instantiated again"""
        args = ()
        kwargs = {}

        if hasattr(val, "__getnewargs_ex__"):
            args, kwargs = val.__getnewargs_ex__()

        elif hasattr(val, "__getnewargs__"):
            args = val.__getnewargs__()

        elif hasattr(val, "__getinitargs__"):
            args = val.__getinitargs__()

        return args, kwargs

    def set_state(self, val, state):
        """Populate val with state (the contents of a previous version of val's
        __dict__ property)"""
        if hasattr(val, "__setstate__"):
            val.__setstate__(state)

        else:
            for k, v in state.items():
                setattr(val, k, v)

    def get_state(self, val):
        """Get the state (the contents of val.__dict__) of val as a dict"""
        if hasattr(val, "__getstate__"):
            d = val.__getstate__()

        else:
            d = val.__dict__

        return d




#!/usr/bin/env python
# http://docs.python.org/distutils/setupscript.html
# http://docs.python.org/2/distutils/examples.html
# https://packaging.python.org/
# https://packaging.python.org/tutorials/distributing-packages/
from setuptools import setup, find_packages
import re
import os
from codecs import open


name = "datatypes"
kwargs = dict(
    name=name,
    description='Utility Classes and Functions that are handy across multiple projects',
    keywords="",
    author='Jay Marcyes',
    author_email='jay@marcyes.com',
    url='http://github.com/Jaymon/{}'.format(name),
    license="MIT",
    classifiers=[ # https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
)

kwargs["tests_require"] = []
kwargs["install_requires"] = []
#kwargs["extras_require"] = {"extra_name": []}


def read(path):
    if os.path.isfile(path):
        with open(path, encoding='utf-8') as f:
            return f.read()
    return ""


vpath = os.path.join(name, "__init__.py")
if os.path.isfile(vpath):
    kwargs["packages"] = find_packages(exclude=["tests", "tests.*", "*_test*", "example*"])

    dpath = os.path.join(name, "data")
    if os.path.isdir(dpath):
        # https://docs.python.org/3/distutils/setupscript.html#installing-package-data
        kwargs["package_data"] = {name: ['data/*']} 

else:
    vpath = "{}.py".format(name)
    kwargs["py_modules"] = [name]

kwargs["version"] = re.search(r"^__version__\s*=\s*[\'\"]([^\'\"]+)", read(vpath), flags=re.I | re.M).group(1)

# https://pypi.org/help/#description-content-type
kwargs["long_description"] = read('README.md')
kwargs["long_description_content_type"] = "text/markdown"


setup(**kwargs)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import heapq
import itertools
import bisect

from .compat import *


class Pool(dict):
    """Generic pool of some values bounded by size, this means when size is reached
    then the least used item will be silently dropped from the pool.

    This class gets interesting when you extend it and add __missing__() so you
    can use this as kind of a hotcache of the maxsize most used items in the Pool

    :Example:
        class HotPool(Pool):
            def __missing__(self, key):
                value = get_some_value_here(key)
                self[key] = value # save the found value at key so it will cache
                return value
    """
    def __init__(self, maxsize=0):
        super(Pool, self).__init__()
        self.pq = PriorityQueue(maxsize, key=lambda value: value)

    def __getitem__(self, key):
        shuffle = key in self
        value = super(Pool, self).__getitem__(key)

        # since this is being accessed again, we move it to the end
        if shuffle:
            self.pq.put(key)

        return value

    def __setitem__(self, key, value):
        super(Pool, self).__setitem__(key, value)
        try:
            self.pq.put(key)

        except OverflowError:
            self.popitem()
            self.pq.put(key)

    def popitem(self):
        dead_val, dead_key, dead_priority = self.pq.popitem()
        del self[dead_key]
        return dead_key, dead_val

    def pop(self, key, *args):
        if key in self:
            self.pq.remove(key)
        return super(Pool, self).pop(key, *args)

    def setdefault(self, key, value):
        if key not in self:
            self[key] = value

    def get(self, key, default=None):
        value = default
        if key in self:
            value = self[key]
        return value

    def update(self, d):
        for k, v in d.items():
            self[k] = v


class PriorityQueue(object):
    """A generic priority queue

    if passing in a tuple (priority, value) then...

        * MinQueue: priority=lambda x: x[0]
        * MaxQueue: priority=lambda x: -x[0]

    Inspiration:

        * https://stackoverflow.com/a/3311765/5006
        * https://docs.python.org/2/library/heapq.html#priority-queue-implementation-notes

    https://en.wikipedia.org/wiki/Priority_queue

    :Example:
        # MinQueue example
        pq = PriorityQueue(priority=lambda x: x[0])
        pq.put((30, "che"))
        pq.put((1, "foo"))
        pq.put((4, "bar"))
        pq.get() # (1, "foo")

    This uses .put() and .get() because that is the same interface as Python's built-in
    queue class.

    If you never pass in priorities it defaults to a FIFO queue, this allows you
    to pass in keys to .put() to set uniqueness and move items back to the top of
    the queue if they have the same key, another name for this might be RefreshQueue
    since a value with the same key will move to the bottom of the queue
    """
    def __init__(self, maxsize=0, key=None, priority=None):
        """create an instance

        :param maxsize: int, the size you want the key to be, 0 for unlimited
        :param key: callable, a callback that will be passed value on every
            call to .put() that doesn't have a key passed in
        :param priority: callable, a callback that will be passed value on every
            call to .put() that doesn't have a priority passed in
        """
        self.pq = []
        self.item_finder = {}
        self.key_counter = itertools.count()
        self.priority_counter = itertools.count()
        self.maxsize = maxsize
        self.removed_count = 0

        if key:
            self.key = key

        if priority:
            self.priority = priority

    def key(self, value):
        """If key isn't passed into the constructor then this method will be called"""
        return next(self.key_counter)

    def priority(self, value):
        """If priority isn't passed into the constructor then this method will be called"""
        return next(self.priority_counter)

    def push(self, value, key=None, priority=None):
        """Same interface as .put() but will remove the next element if the queue
        is full so value can be placed into the queue"""
        try:
            self.put(value, key=key, priority=priority)

        except OverflowError:
            self.popitem()
            self.put(value, key=key, priority=priority)

    def put(self, value, key=None, priority=None):
        """add a value to the queue with priority, using the key to know uniqueness

        :param value: mixed, the value to add to the queue
        :param key: string|int, this is used to determine uniqueness in the queue,
            if key is already in the queue, then the val will be replaced in the
            queue with the new priority, if key is None then .key(value) will be
            called to determine a key for value
        :param priority: int, the priority of value, if None then .priority(value)
            will be called to determine a priority for the value, defaults to FIFO
        """
        deleted = False

        if key is None:
            key = self.key(value)

        if priority is None:
            priority = self.priority(value)

        if key in self.item_finder:
            self.remove(key)

        else:
            # keep the queue contained
            if self.full():
                raise OverflowError("Queue is full")

        item = [priority, key, value, deleted]
        self.item_finder[key] = item
        heapq.heappush(self.pq, item)

    def remove(self, key):
        """remove the value found at key from the queue"""
        item = self.item_finder.pop(key)
        item[-1] = True
        self.removed_count += 1

    def popitem(self):
        """remove the next prioritized (value, key, priority) and return it"""
        pq = self.pq
        while pq:
            priority, key, value, deleted = heapq.heappop(pq)
            if deleted:
                self.removed_count -= 1

            else:
                del self.item_finder[key]
                return value, key, priority

        raise KeyError("Pop from an empty queue")

    def get(self):
        """Remove the next prioritized val and return it"""
        value, key, priority = self.popitem()
        return value

    def full(self):
        """Return True if the queue is full"""
        if not self.maxsize: return False
        return len(self.pq) >= (self.maxsize + self.removed_count)

    def keys(self):
        """return the keys in the order they are in the queue"""
        for x in self.pq:
            if not x[3]:
                yield x[1]

    def values(self):
        """return the values in the order they are in the queue"""
        for x in self.pq:
            if not x[3]:
                yield x[2]

    def qsize(self):
        """for similarity to python's queue interface"""
        return self.maxsize

    def __len__(self):
        return len(self.item_finder)

    def __bool__(self):
        return bool(len(self))
    __nonzero__ = __bool__


class NormalizeDict(dict):
    """A normalizing dictionary, taken from herd.utils.NormalizeDict"""
    def __init__(self, *args, **kwargs):
        super(NormalizeDict, self).__init__()
        self.update(*args, **kwargs)

    def __setitem__(self, k, v):
        k = self.normalize_key(k)
        v = self.normalize_value(v)
        return super(NormalizeDict, self).__setitem__(k, v)

    def __delitem__(self, k):
        k = self.normalize_key(k)
        return super(NormalizeDict, self).__delitem__(k)

    def __getitem__(self, k):
        k = self.normalize_key(k)
        return super(NormalizeDict, self).__getitem__(k)

    def __contains__(self, k):
        k = self.normalize_key(k)
        return super(NormalizeDict, self).__contains__(k)

    def setdefault(self, k, default=None):
        k = self.normalize_key(k)
        v = self.normalize_value(default)
        return super(NormalizeDict, self).setdefault(k, v)

    def update(self, *args, **kwargs):
        # create temp dictionary so I don't have to mess with the arguments
        d = dict(*args, **kwargs)
        for k, v in d.items():
            self[k] = v

    def pop(self, k, default=None):
        k = self.normalize_key(k)
        v = self.normalize_value(default)
        return super(NormalizeDict, self).pop(k, v)

    def get(self, k, default=None):
        k = self.normalize_key(k)
        v = self.normalize_value(default)
        return super(NormalizeDict, self).get(k, v)

    def normalize_key(self, k):
        return k

    def normalize_value(self, v):
        return v


class idict(NormalizeDict):
    """A case insensitive dictionary, adapted from herd.utils.NormalizeDict, naming
    convention is meant to mimic python's use of i* for case-insensitive functions and
    python's built-in dict class (which is why the classname is all lowercase)"""
    def __init__(self, *args, **kwargs):
        self.ikeys = {} # hold mappings from normalized keys to the actual key
        super(idict, self).__init__(*args, **kwargs)

    def normalize_key(self, k):
        nk = k.lower()
        if nk in self.ikeys:
            k = self.ikeys[nk]
        self.ikeys[nk] = k
        return k
Idict = idict
IDict = idict
iDict = idict


class Trie(object):
    """https://en.wikipedia.org/wiki/Trie"""
    def __init__(self, *values):
        self.values = {} #defaultdict(list)

        for value in values:
            self.add(value)

    def add(self, value):
        if value:
            ch = value[0]
            remainder = value[1:]
            if ch not in self.values:
                self.values[ch] = None

            if remainder:
                if self.values[ch] is None:
                    self.values[ch] = type(self)(remainder)

                else:
                    self.values[ch].add(self.normalize_value(remainder))

    def has(self, value):
        ret = True
        ch = self.normalize_value(value[0])
        remainder = value[1:]
        if ch in self.values:
            if remainder:
                if self.values[ch] is None:
                    ret = False
                else:
                    ret = self.values[ch].has(remainder)

        else:
            ret = False

        #pout.v(ch, ret, self.values.keys())
        return ret

    def __contains__(self, value):
        return self.has(value)

    def normalize_value(self, value):
        return value.lower()


# class FrozenList(list): # a read only list that can only be set via constructor

class AppendList(list):
    """A READ ONLY list that does all adding of items through the append method
    To customize this list just extend the append() method

    https://docs.python.org/3/tutorial/datastructures.html#more-on-lists
    """
    def __init__(self, iterable=None):
        """
        https://docs.python.org/3/library/stdtypes.html?highlight=list#list
        """
        if iterable:
            self.extend(iterable)

    def append(self, x):
        return super(AppendList, self).append(x)

    def extend(self, iterable):
        for x in iterable:
            self.append(x)

    def __setitem__(self, *args, **kwargs):
        raise NotImplementedError()
    insert = __setitem__
    remove = __setitem__
    pop = __setitem__
    clear = __setitem__
    __delitem__ = __setitem__


class OrderedList(list):
    """Keep a list sorted as you append or extend it

    An ordered list, this sorts items from smallest to largest using key, so
    if you want MaxQueue like functionality use negative values: .pop(-1) and
    if you want MinQueue like functionality use positive values: .pop(0)
    """
    def __init__(self, iterable=None, key=None):
        if key:
            self.key = key
        self._keys = []
        super(OrderedList, self).__init__()
        if iterable:
            for x in iterable:
                self.append(x)

    def key(self, x):
        return x

    def append(self, x):
        k = self.key(x)
        # https://docs.python.org/3/library/bisect.html#bisect.bisect_right
        i = bisect.bisect_right(self._keys, k)
        if i is None:
            super(OrderedList, self).append((self.key(x), x))
            self._keys.append(k)
        else:
            super(OrderedList, self).insert(i, (self.key(x), x))
            self._keys.insert(i, k)

    def extend(self, iterable):
        for x in iterable:
            self.append(x)

    def remove(self, x):
        k = self.key(x)
        self._keys.remove(k)
        super(OrderedList, self).remove((k, x))

    def pop(self, i=-1):
        self._keys.pop(i)
        return super(OrderedList, self).pop(i)[-1]

    def clear(self):
        super(OrderedList, self).clear()
        self._keys.clear()

    def __iter__(self):
        for x in super(OrderedList, self).__iter__():
            yield x[-1]

    def __getitem__(self, i):
        return super(OrderedList, self).__getitem__(i)[-1]

    def insert(self, i, x):
        raise NotImplementedError()
    def __setitem__(self, x):
        raise NotImplementedError()
    def reverse(self):
        raise NotImplementedError()
    def sort(self):
        raise NotImplementedError()


class MembershipSet(set):
    """A set with all the AND, OR, and UNION operations disabled, making it really
    only handy for testing membership

    This is really more of a skeleton for the few times I've had to do this in a
    project, usually we are implementing custom functionality that acts like a set
    and so it will be nice to just be able to extend this and not have to worry
    about disabling the unsupported methods

    I thought of the name HotSet also

    If you need a readonly set, use frozenset:
        https://docs.python.org/3/library/stdtypes.html#frozenset
    """
    def __init__(self, iterable=None):
        if not iterable:
            iterable = []

        super(MembershipSet, self).__init__(iterable)

    def add(self, elem):
        super(MembershipSet, self).add(perm)

    def remove(self, elem):
        super(MembershipSet, self).remove(perm)

    def discard(self, elem):
        try:
            self.remove(elem)
        except KeyError:
            pass

    def clear(self):
        super(MembershipSet, self).clear()

    def update(self, *others):
        for iterable in others:
            super(MembershipSet, self).update(iterable)

    def noimp(self, *args, **kwargs):
        raise NotImplementedError()

    pop = noimp
    __sub__ = noimp
    __and__ = noimp
    __or__ = noimp
    __xor__ = noimp
    __isub__ = noimp
    __iand__ = noimp
    __ior__ = noimp
    __ixor__ = noimp
    intersection_update = noimp
    difference_update = noimp
    symmetric_difference_update = noimp
    symmetric_difference = noimp
    difference = noimp
    intersection = noimp
    union = noimp = noimp


class ListIterator(list):
    """A base interface for iterators that should quack like a builtin list

    it defines the list interface as much as possible

    Taken from prom.query.BaseIterator on 12-15-2020

    list interface methods:
        https://docs.python.org/3/tutorial/datastructures.html#more-on-lists

    http://docs.python.org/2/library/stdtypes.html#iterator-types
    """
    def __init__(self, *args, **kwargs):
        super(Iterator, self).__init__(*args, **kwargs)

    def next(self):
        raise NotImplementedError()

    def __next__(self):
        """needed for py3 api compatibility"""
        return self.next()

    def __iter__(self):
        return self

    def __nonzero__(self):
        return True if self.count() else False

    def __len__(self):
        return self.count()

    def count(self):
        """list interface compatibility"""
        raise NotImplementedError()

    def __getslice__(self, i, j):
        """required for slicing in python 2 when extending built-in types like list

        https://docs.python.org/2/reference/datamodel.html#object.__getslice__
        https://stackoverflow.com/questions/2936863/implementing-slicing-in-getitem#comment39878974_2936876
        """
        return self.__getitem__(slice(i, j))

    def __getitem__(self, k):
        raise NotImplementedError()

    def reverse(self):
        """list interface compatibility"""
        raise NotImplementedError()

    def __reversed__(self):
        it = self.copy()
        it.reverse()
        return it

    def sort(self, *args, **kwargs):
        """list interface compatibility"""
        raise NotImplementedError()

    def __getattr__(self, k):
        """
        this allows you to focus in on certain fields of results

        It's just an easier way of doing: (getattr(x, k, None) for x in self)
        """
        raise NotImplementedError()

    def append(self, x):
        """list interface compatibility"""
        raise NotImplementedError()

    def extend(self, iterable):
        """list interface compatibility"""
        raise NotImplementedError()

    def remove(self, x):
        """list interface compatibility"""
        raise NotImplementedError()

    def pop(self, i=-1):
        """list interface compatibility"""
        raise NotImplementedError()

    def clear(self):
        """list interface compatibility"""
        raise NotImplementedError()

    def index(self, x, start=None, end=None):
        """list interface compatibility"""
        raise NotImplementedError()

    def copy(self):
        """list interface compatibility"""
        raise NotImplementedError()




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import subprocess
import os
import inspect
import sys

from .path import Dirpath


# 9-20-2017 -- I'm still testing these with some projects trying to get
#   the api right and so they aren't fully integrated yet
# 3-29-2018 - I've further integrated this with the testdata.start_service() and
#   testdata.stop_service() methods
# 3-24-2020 - Added Systemd support
# 8-4-2020 - moved this from testdata to datatypes (though it will also be in
#   testdata for a while, this should be considered the DRY master version), this
#   basically converts testdata.start_service() and testdata.stop_service() into
#   Service()
# 3-30-2021 - Removed testdata Service code in favor of this implementation, and 
#   made the code a little easier to follow but also a little more magical (eg, 
#   the Service class now returns other BaseService instances)


class BaseService(object):
    """base class for services"""

    sudo = True
    """If true then sudo should be added to the command"""

    ignore_failure = True
    """If True then failures when running the command will be ignored, failure is
    usually defined as an exit code >0"""

    @property
    def path(self):
        raise NotImplementedError()

    def __init__(self, name, ignore_failure=True, sudo=True):
        self.name = name
        self.ignore_failure = ignore_failure
        self.sudo = sudo

    def format_cmd(self, action, **kwargs):
        cmd = []
        if self.sudo:
            cmd.append("sudo")

        c, kw = self._format_cmd(action, **kwargs)
        cmd.extend(c)
        kwargs.update(kw)
        return cmd, kwargs

    def _format_cmd(self, action, **kwargs):
        raise NotImplementedError()

    def is_running(self):
        raise NotImplementedError()

    def start(self):
        cmd, kwargs = self.format_cmd("start")
        self.run(cmd, kwargs)

    def restart(self):
        self.stop()
        self.start()

    def stop(self):
        cmd, kwargs = self.format_cmd("stop")
        self.run(cmd, kwargs)

    def status(self):
        cmd, kwargs = self.format_cmd("status")
        return self.run(cmd, kwargs)

    def run(self, cmd, kwargs):
        try:
            ret = subprocess.check_output(cmd, **kwargs)

        except subprocess.CalledProcessError as e:
            if self.ignore_failure:
                ret = None
            else:
                raise

        return ret

    def exists(self):
        path = self.path
        return os.path.isfile(path) if path else False


class Upstart(BaseService):
    """Handle starting Upstart services"""
    @property
    def path(self):
        return "/etc/init/{}".format(self.name)

    def _format_cmd(self, action, **kwargs):
        return [action, self.name], kwargs

    def is_running(self):
        return "start/running" in self.status()


class InitD(BaseService):
    """Handle starting init.d services"""
    @property
    def path(self):
        return "/etc/init.d/{}".format(self.name)

    def _format_cmd(self, action, **kwargs):
        return [self.path, action], kwargs

    def is_running(self):
        return self.status() is None


class Systemd(BaseService):
    """Handle starting Systemd services"""
    @property
    def path(self):
        # https://unix.stackexchange.com/a/367237/118750
        ret = getattr(self, "_path", None)
        if ret is None:
            dirs = [
                "/etc/systemd/system",
                "/etc/systemd/user",
                "/usr/local/lib/systemd/system",
                "/usr/lib/systemd/system",
                "/usr/lib/systemd/user",
                "/usr/local/lib/systemd/user",
                "/run/systemd/system",
                "/run/systemd/user",
            ]

            if "XDG_CONFIG_HOME" in os.environ:
                dirs.append(os.path.join(os.environ["XDG_CONFIG_HOME"], "systemd", "user"))
            else:
                dirs.append("~/.config/systemd/user")

            if "XDG_RUNTIME_DIR" in os.environ:
                dirs.append(os.path.join(os.environ["XDG_RUNTIME_DIR"], "systemd", "user"))

            if "XDG_DATA_HOME" in os.environ:
                dirs.append(os.path.join(os.environ["XDG_DATA_HOME"], "systemd", "user"))
            else:
                dirs.append("~/.local/share/systemd/user")

            # go through all the possible locations for systemd unit files and find name
            name = self.name
            if "." not in name:
                name = "{}.".format(name)
            for path in dirs:
                d = Dirpath(path)
                if d.exists():
                    for f in d.files():
                        if f.basename.startswith(name):
                            ret = f.path
                            self._path = ret
                            break

                if ret:
                    break

        return ret

    def _format_cmd(self, action, **kwargs):
        return ["systemctl", action, self.name], kwargs

    def is_running(self):
        return "Active: active (running)" in self.status()


class Service(BaseService):
    """Catch-all to provide a common interface for any of the other services, it
    will check all the services in the order returned by .service_classes() to find
    the correct type of service and set it in .service_class so you can start/stop
    the service without worrying about the underlying service type (eg, Systemd, Upstart)

    :Example:
        s = Service("postgresql")
        s.restart()
    """
    @classmethod
    def service_classes(cls):
        """Return all the BaseService subclasses that should be checked and the
        order they should be checked in

        :returns: list, the BaseService children to check for a service, in priority
            order
        """
        service_classes = []
        for name, o in inspect.getmembers(sys.modules[__name__]):
            try:
                if issubclass(o, BaseService) and o is not BaseService and not issubclass(o, Service):
                    service_classes.append(o)

            except TypeError:
                pass

        return service_classes

    def __new__(cls, name, *args, **kwargs):
        """Magic new that will return the BaseService instance that matches name"""
        for service_class in cls.service_classes():
            s = service_class(name, *args, **kwargs)
            if s.exists():
                return s

        raise RuntimeError("Could not find a valid service for {}".format(name))

